/*
** Module   : ColorMapLoader.java
** Abstract : Implements loading of color-map configuration.
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date--    ----------------Description-----------------
** 001 HC  20140119    Created initial version.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;

import org.apache.commons.logging.*;
import org.dom4j.*;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.dom4j.tree.DefaultElement;
import org.xml.sax.*;


/**
 * <p>
 * The class is responsible for loading the color-map configuration. The color-map
 * configuration provides the source of data for the Progress COLOR protermcap-attribute. 
 * </p>
 * <p>
 * The configuration is stored in the color-map.xml classpath resource stored at the root 
 * of the namespace.
 * </p> 
 */
public class ColorMapLoader
{
   private static final Log LOG = LogFactory.getLog(ColorMapLoader.class);
   
   /**
    * the resource path to he color map XML 
    */
   private static final String COLOR_MAP_XML_RESOURCE = "/color-map.xml";

   /**
    * the resource path to he color map DTD
    */
   private static final String COLOR_MAP_DTD_RESOURCE = "/color-map.dtd";
   
   /**
    * the DTD id reference used in the color map XML
    */
   private static final String COLOR_MAP_DTD_ID = "color-map.dtd";
   
   /**
    * 'types' color-map XML element name
    */
   private static final String E_NAME_TYPES = "types";

   /**
    * 'index' color-map XML attribute name
    */
   private static final String E_NAME_INDEX = "index";
   
   /**
    * 'name' color-map XML attribute name
    */
   private static final String E_NAME_NAME = "name";

   /**
    * 'foreground' color-map XML attribute name
    */
   private static final String E_NAME_FOREGROUND = "foreground";

   /**
    * 'background' color-map XML attribute name
    */
   private static final String E_NAME_BACKGROUND = "background";

   /**
    * 'attributes' color-map XML attribute name
    */
   private static final String E_NAME_ATTRIBUTES = "attributes";

   
   /**
    * Loads the color map configuration data into the map of terminal names
    * to map of color name to color values.
    * 
    * Returns a map of terminal names to the color maps. I.e. each terminal 
    * has its own color map. The key to the color map is either the color name
    * (String) or the color index (Integer) (see color-map.xml).
    * 
    * @return   The map of color maps. See above.
    * 
    */
   public Map<String, Map<Object, Color>> loadAllMappings()
   {
      // get the document
      Document document = parseDocument();
      if (document == null)
      {
         return null;
      }

      Map<String, Map<Object, Color>> output = new HashMap<String, Map<Object, Color>>();

      // iterate the root elements
      for (Iterator<?> i = document.getRootElement().elementIterator(); i.hasNext();)
      {
         Object element = i.next();

         if (!(element instanceof Element))
         {
            continue;
         }

         // the ansi-terminal level
         Element ansiTerminal = (Element) element;
         String types = ansiTerminal.attribute(E_NAME_TYPES).getText();
         List<?> colors = ansiTerminal.elements();

         // iterate the colors
         Map<Object, Color> colorMap = new HashMap<Object, Color>();
         for (int k = 0; k < colors.size(); k++)
         {
            if (!(colors.get(k) instanceof DefaultElement))
            {
               continue;
            }

            DefaultElement colorElement = (DefaultElement) colors.get(k);
            
            Integer index = null;
            try 
            {
               String indexString = colorElement.attributeValue(E_NAME_INDEX);
               index = Integer.valueOf(indexString);
            }
            catch (NumberFormatException ex)
            {
               // skip the color with invalid index
               continue;
            }
            
            String name = colorElement.attributeValue(E_NAME_NAME);
            Color color = parseColor(colorElement);
            colorMap.put(name, color);
            colorMap.put(index, color);
         }

         // type attribute is a comma-separated list
         for (String type : types.split(","))
         {
            output.put(type.trim(), colorMap);
         }
      }

      return output;
   }

   /**
    * Converts the DOM instance of the 'color' xml element into
    * the instance of type {@link Color}.
    * 
    * @param   colorElement
    *          DOM instance representing the 'color' element.
    * 
    * @return  The instance of type {@link Color}
    */
   private Color parseColor(DefaultElement colorElement)
   {
      String foreground = colorElement.attributeValue(E_NAME_FOREGROUND).trim().toUpperCase();
      String background = colorElement.attributeValue(E_NAME_BACKGROUND).trim().toUpperCase();
      String attributes = colorElement.attributeValue(E_NAME_ATTRIBUTES);

      // 'attributes' is a comma-separated list of color attributes
      int attribute = attributes == null ? 0 : parseAttributes(attributes.split(","));
      ColorPalette parsedForeground = ColorPalette.valueOf(foreground);
      ColorPalette parsedBackround = ColorPalette.valueOf(background);

      return new Color(attribute, Color.toColorValue(parsedForeground, parsedBackround));
   }

   /**
    * Converts the array of named color attributes into
    * a bitset compatible with {@link ColorAttribute}. 
    * the instance of type {@link Color}.
    * 
    * @param   attributes
    *          Array of named color attributes.
    * 
    * @return  Bitset compatible with {@link ColorAttribute}.
    */
   private int parseAttributes(String[] attributes)
   {
      int attribute = 0;

      for (String attributeName : attributes)
      {
         try
         {
            attribute |= ColorAttribute.valueOf(attributeName.trim().toUpperCase()).getValue();
         } 
         catch (IllegalArgumentException ex)
         {
            // ignore the invalid attribute names
         }
      }

      return attribute;
   }

   /**
    * Loads the color-map xml as classpath resource and using
    * dom4j sax parser builds its DOM representation.
    *  
    * @return  the DOM representation of the color-map xml.
    */
   private Document parseDocument()
   {
      try
      {
         // the resolver returns the DTD used by the sax parser
         // to validate the xml
         EntityResolver resolver = new EntityResolver()
         {
            public InputSource resolveEntity(String publicId, String systemId)
            {
               if (systemId.endsWith(COLOR_MAP_DTD_ID))
               {
                  // get the xml from the classpath resource
                  InputStream in = getClass().getResourceAsStream(COLOR_MAP_DTD_RESOURCE);
                  if (in == null)
                  {
                     // warn, now this resource is expected when color-map.xml was found
                     LOG.warn("Color map DTD resource not found at " + COLOR_MAP_DTD_RESOURCE);
                     return null;
                  }
                  return new InputSource(in);
               }
               return null;
            }
         };

         // get the xml from the classpath resource 
         InputStream inputStream = ClassLoader.class.getResourceAsStream(COLOR_MAP_XML_RESOURCE);
         if (inputStream == null)
         {
            LOG.info("Color map XML resource not found at " + COLOR_MAP_XML_RESOURCE);
            return null;
         }
         
         SAXReader reader = new SAXReader(true);
         reader.setEntityResolver(resolver);
         Document document = reader.read(inputStream);
         return document;
      } 
      catch (DocumentException e)
      {
         LOG.error("Failed to load the color map XML", e);
         return null;
      }
   }
}
